/*
 * File: messages.c
 * Date: 2011-11-14
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */

/***

The functions in this file handle the public usage of the message line.
Functions used by the message line itself are in input.c.

***/

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"rho.h"
#include	"messages.h"
#include	"bind.h"
#include	"buffer.h"
#include	"display.h"
#include	"exec.h"
#include	"terminal.h"
#include	"input.h"
#include	"utils.h"


/*++

Output a string of input characters.

++*/
static void outstring(s)
char *s;                        /* string to output */

{
    if (disinp)
        while (*s)
            mlout(*s++);
}


/*++

Pause to see stdout

++*/
void ml_end()
{
    mlwrite("[End]");
    ttflush();
    ttgetc();
}


/*++

A more generalized prompt/reply function allowing the caller to specify
the proper terminator. If the terminator is not a return('\r'), return
will echo as "<NL>"

++*/
int mlgetstr(buf, nbuf, eolchar)
char *buf;
int nbuf;
int eolchar;

{
    int cpos;                   /* current character position in string */
    int c;                      /* current input character */
    int ec;                     /* extended current input character */
    int quotef;                 /* are we quoting the next char? */
    char *kp;                   /* pointer into key_name */
    char key_name[10];          /* name of a quoted key */

    cpos = 0;
    quotef = FALSE;

    for (;;) {
        /* get a character from the user */
        ec = mlgetkey();

        /* if they hit the line terminate, wrap it up */
        if (ec == eolchar && quotef == FALSE) {
            buf[cpos++] = 0;

            /* clear the message line */
            mlerase();

            /* if we default the buffer, return FALSE */
            if (buf[0] == 0)
                return (FALSE);

            return (TRUE);
        }

        /* change from command form back to character form */
        c = ectoc(ec);

        if ((ec == abortc) && (quotef == FALSE)) {
            /* Abort the input? */
            abort_command(FALSE, 0);
            TTflush();
            return (ABORT);
        }

        /* if it is from a function key, blow it off */
        if ((quotef == FALSE) && (ec & SPEC))
            continue;

        /* rubout/erase */
        if ((c == 0x7F || c == 0x08) && quotef == FALSE) {
            if (cpos != 0) {
                outstring("\b \b");
                --ttcol;

                if (buf[--cpos] < 0x20) {
                    outstring("\b \b");
                    --ttcol;
                }

                if (buf[cpos] == '\r') {
                    outstring("\b\b  \b\b");
                    ttcol -= 2;
                }
                TTflush();
            }
            continue;
        }

        /* C-K, kill default buffer and return null */
        if (c == 0x0b && quotef == FALSE) {

            /* clear the buffer */
            buf[0] = 0;

            /* clear the message line and return */
            mlwrite("");
            TTflush();
            return (TRUE);
        }

        /* C-U, kill */
        if (c == 0x15 && quotef == FALSE) {

            while (cpos != 0) {
                outstring("\b \b");
                --ttcol;

                if (buf[--cpos] < 0x20) {
                    outstring("\b \b");
                    --ttcol;
                }
                if (buf[cpos] == '\r') {
                    outstring("\b\b  \b\b");
                    ttcol -= 2;
                }
            }
            TTflush();
            continue;
        }

        /* quoting next character? */
        if ((ec == quotec) && (quotef == FALSE)) {
            quotef = TRUE;
            continue;
        }

        quotef = FALSE;

        /* if it is from a function key,
           insert it's name since it was quoted */
        if (ec & SPEC) {
            (void) cmd2str(ec, key_name);
            kp = key_name;
            while (*kp) {
                if (cpos < nbuf - 1) {
                    if (disinp)
                        mlout(*kp);
                    buf[cpos++] = *kp++;
                    ++ttcol;
                }
            }
            TTflush();
            continue;
        }

        /* insert the character in the string! */
        if (cpos < nbuf - 1) {

            buf[cpos++] = c;

            if ((c < ' ') && (c != '\r')) {
                outstring("^");
                ++ttcol;
                c ^= 0x40;
            }

            if (c != '\r') {
                if (disinp)
                    mlout(c);
            } else {            /* put out <NL> for <ret> */
                outstring("<NL>");
                ttcol += 3;
            }
            ++ttcol;
            TTflush();
        }
    }
}



/*++

Get one keystroke. The legal prefixs here are the SPEC and CTRL prefixes.

++*/
int mlgetkey()
{
    int c;                      /* next input character */
    int upper;                  /* upper byte of the extended sequence */

    /* get a keystroke */
    c = tgetc();

    /* if it exists, process an escape sequence */
    if (c == 0) {

        /* get the event type */
        upper = tgetc();

        /* get the event code */
        c = tgetc();

        /* if it is a function key... map it */
        c = (upper << 8) | c;
    }

    /* yank out the control prefix */
    if (((c & 255) >= 0x00 && (c & 255) <= 0x1F) || (c & 255) == 0x7F)
        c = CTRL | (c ^ 0x40);

    /* return the character */
    return (c);
}





/*++

Write out an integer, in the specified radix. Update the physical cursor
position.

++*/
static void mlputi(i, r)
int i, r;

{
    int q;
    static char hexdigits[] = "0123456789ABCDEF";

    if (i < 0) {
        i = -i;
        mlout('-');
    }

    q = i / r;

    if (q != 0)
        mlputi(q, r);

    mlout(hexdigits[i % r]);
    ++ttcol;
}


/*++

Does the same as mlputi except as a long integer.

++*/
static void mlputli(l, r)
long l;
int r;

{
    long q;

    if (l < 0) {
        l = -l;
        mlout('-');
    }

    q = l / r;

    if (q != 0)
        mlputli(q, r);

    mlout((int) (l % r) + '0');
    ++ttcol;
}


/*++

Write out a scaled integer with two decimal places.

++*/
static void mlputf(s)
int s;                          /* scaled integer to output */

{
    int i;                      /* integer portion of number */
    int f;                      /* fractional portion of number */

    /* break it up */
    i = s / 100;
    f = s % 100;

    /* send out the integer portion */
    mlputi(i, 10);
    mlout('.');
    mlout((f / 10) + '0');
    mlout((f % 10) + '0');
    ttcol += 3;
}





/*++

Get a filename from the user.

++*/
char *prompt4filename(prompt)
char *prompt;                   /* prompt to user on message line */

{
    char *sp;                   /* ptr to the returned string */

    /* get a file name, default to current buffer's */
    if (curbp && curbp->b_name && strcmp(curbp->b_name, "") != 0)
        sp = complete(prompt, curbp->b_name, CMP_FILENAME, LENPATH);
    else
        sp = complete(prompt, NULL, CMP_FILENAME, LENPATH);

    return (sp);
}




/*++

Get a buffer's name from the user.

++*/
BUFFER *prompt4buffer(prompt, defval, createflag)
char *prompt;                   /* prompt to user on command line */
char *defval;                   /* default value to display to user */
int createflag;                 /* should this create a new buffer? */

{
    char *sp;                   /* ptr to the returned string */

    sp = complete(prompt, defval, CMP_BUFFER, LENPATH);
    if (sp == NULL)
        return (NULL);

    return (exist_buffer(sp, createflag, 0));
}


/*++

Get a command name (rho_name.h) from the command line. Command completion
means that pressing a <SPACE> will attempt to complete an unfinished
command name if it is unique.

++**/
int (*prompt4command(prompt)) ()
char *prompt;                   /* string to prompt with */

{
    char *sp;                   /* ptr to the returned string */

    sp = complete(prompt, NULL, CMP_COMMAND, NSTRING);
    if (sp == NULL)
        return (NULL);

    return (get_fn_name(sp));
}


  /*  start msgline - msgiface analysis here  */


/*++

This erase is only in main.editloop.

++*/
void mlferase()
{
    int save_discmd;

    save_discmd = discmd;
    discmd = TRUE;
    mlerase();
    discmd = save_discmd;;
}


/*++

Erase the message line. This is a special routine because the message line
is not considered to be part of the virtual screen. It always works
immediately; the terminal buffer is flushed via a call to the flusher.

++*/
void mlerase()
{
    int i;

    movecursor(term.t_nrow, 0);
    if (discmd == FALSE)
        return;

    if (eolexist == TRUE)
        TTeeol();
    else {
        for (i = 0; i < term.t_ncol - 1; i++)
            TTputc(' ');

        /* force the move! */
/*		movecursor(term.t_nrow, 1);*/
        movecursor(term.t_nrow, 0);
    }
    TTflush();
    mpresf = FALSE;
}


/*++

Write a single character to the messageline.  Used for
building messages.

++*/
void mlout(c)
int c;                          /* character to write */

{
    if (ttcol + 1 < term.t_ncol)
        TTputc(c);
    if (c != '\b')
        *lastptr++ = c;
    else if (lastptr > &lastmesg[0])
        --lastptr;
}



/*++

Write a message into the message line. Keep track of the physical cursor
position. A small class of printf like format items is handled. Assumes
the stack grows down; this assumption is made by the "+=" in the argument
scan loop. Set the "message line" flag TRUE.  Don't write beyond the
end of the current terminal width.

++*/
void mlwrite(char *fmt, ...)
/* char * fmt;
   variable argument list
arg1 = format string
arg2+ = arguments in that string */
{
    int c;                      /* current char in format string */
    va_list ap;                 /* ptr to current data field */
    int arg_int;                /* integer argument */
    long arg_long;              /* long argument */
    char *arg_str;              /* string argument */

    /* if we are not currently echoing on the command line, abort this */
    if (discmd == FALSE)
        return;

    /* point to the first argument */
    va_start(ap, fmt);

    /* if we can not erase to end-of-line, do it manually */
    if (eolexist == FALSE) {
        mlerase();
        TTflush();
    }

    movecursor(term.t_nrow, 0);
    lastptr = &lastmesg[0];     /* setup to record message */
    while ((c = *fmt++) != 0) {
        if (c != '%') {
            mlout(c);
            ++ttcol;
        } else {
            c = *fmt++;
            switch (c) {
            case 'd':
                arg_int = va_arg(ap, int);
                mlputi(arg_int, 10);
                break;

            case 'o':
                arg_int = va_arg(ap, int);
                mlputi(arg_int, 8);
                break;

            case 'x':
                arg_int = va_arg(ap, int);
                mlputi(arg_int, 16);
                break;

            case 'D':
                arg_long = va_arg(ap, long);
                mlputli(arg_long, 10);
                break;

            case 's':
                arg_str = va_arg(ap, char *);
                mlputs(arg_str);
                break;

            case 'f':
                arg_int = va_arg(ap, int);
                mlputf(arg_int);
                break;

            default:
                mlout(c);
                ++ttcol;
            }
        }
    }

    /* if we can, erase to the end of screen */
    if (eolexist == TRUE)
        TTeeol();
    TTflush();
    mpresf = TRUE;
    *lastptr = 0;               /* terminate lastmesg[] */
    va_end(ap);
}


/*++

Ask a yes or no question in the message line. Return either TRUE, FALSE, or
ABORT. The ABORT status is returned if the user bumps out of the question
with a ^G. Used any time a confirmation is required.

++*/
int mlyesno(prompt)
char *prompt;

{
    int c;                      /* input character */
    char buf[NPAT];             /* prompt to user */

    for (;;) {
        /* build and prompt the user */
        strcpy(buf, prompt);
        strcat(buf, TEXT162);   /*" [y/n]? " */
        mlwrite(buf);

        c = get_keystrokes();

        if (c == ectoc(abortc))
            return (ABORT);

        if ((c == 'n') || (c == 'N') || (c & (SPEC | CTRL | META | CTLX)))
            return (FALSE);

#if	FRENCH
        if (c == 'o' || c == 'O')
            return (TRUE);
#elif GERMAN
        if (c == 'j' || c == 'J')
            return (TRUE);
#elif SPANISH
        if (c == 's' || c == 'S')
            return (TRUE);
#else
        if (c == 'y' || c == 'Y')
            return (TRUE);
#endif

        return (FALSE);
    }
}


/*++

Write a prompt into the message line, then read back a response. Keep
track of the physical position of the cursor. If we are in a keyboard
macro throw the prompt away, and return the remembered response. This
lets macros run at full speed. The reply is always terminated by a carriage
return. Handle erase, kill, and abort keys.

++*/
int mlreply(prompt, buf, nbuf)
char *prompt;
char *buf;
int nbuf;

{
    return (next_arg(prompt, buf, nbuf, ctoec((int) '\r')));
}



/*++

Force a string out to the message line regardless of the current $discmd
setting. This is needed when $debug is TRUE and for the write-message
and clear-message-line commands.

++*/
void mlforce(s)
char *s;                        /* string to force out */

{
    int oldcmd;                 /* original command display flag */

    oldcmd = discmd;            /* save the discmd value */
    discmd = TRUE;              /* and turn display on */
    mlwrite(s);                 /* write the string out */
    discmd = oldcmd;            /* and restore the original setting */
}



/*++

Display a serious error message. This is replaced by a system-specific
function when a multitasking system that does not like these kind of
errors is used, so that the user can be offered to abort the application

++*/
void mlabort(s)
char *s;
{
    mlforce(s);
}


/*++

Write out a string. Update the physical cursor position. This assumes
that the characters in the string all have width "1"; if this is not
the case things will get screwed up a little.

++*/
void mlputs(s)
char *s;

{
    int c;

    while ((c = *s++) != 0) {
        mlout(c);
        ++ttcol;
    }
}



/*++

Display a prompt [with optional default] and the input terminator.

++*/
int mlprompt(prompt, dflt, iterm)
char *prompt;
char *dflt;
int iterm;
{
    int tcol;
    char buf[NSTRING];

    /* don't bother displaying if we are't currently */
    if (discmd == FALSE)
        return (0);

    /* show the passed in prompt */
    mlwrite(prompt);
    tcol = (int) strlen(prompt);

    /* If there's a default, put it in brackets and show it. */
    if (dflt != NULL && *dflt != '\0') {
        mlout('[');
        tcol = 1 + echostring(dflt, tcol + 1, NPAT / 2);
        mlout(']');
    }

    /* Display the proper current search terminator character. */
    mlout('<');
    switch (iterm) {
    case CTRL | '[':
        mlputs("META");
        tcol += 8;
        break;
    case CTRL | 'M':
        mlputs("NL");
        tcol += 6;
        break;
    default:
        mlputs(cmd2str(iterm, buf));
        tcol += strlen(buf) + 4;
    }
    mlputs(">: ");
    TTflush();
    return (tcol);
}



/*++

Uses echochar() to put out a string.  Checks for NULL.

++*/
int echostring(str, tcol, uptocol)
char *str;                      /* characters to be echoed */
int tcol;                       /* column to be echoed in */
int uptocol;                    /* last column to be echoed in */
{
    if (str != NULL) {
        while (*str) {
            tcol = echo_char(*str++, tcol);
            if (tcol >= uptocol) {
                mlout('$');
                tcol++;
                break;
            }
        }
    }
    return (tcol);
}



/*++

Routine to echo i-search and message-prompting characters.

++*/
int echo_char(c, col)
char c;                         /* character to be echoed */
int col;                        /* column to be echoed in */

{
    movecursor(term.t_nrow, col);       /* Position the cursor  */
    if (c == '\r') {            /* Newline character?   */
        mlout('<');
        mlout('N');
        mlout('L');
        mlout('>');
        col += 3;
    } else if ((c < ' ') || (c == 0x7F)) {      /* Vanilla control char and Rubout:   */
        mlout('^');             /* Yes, output prefix           */
        mlout(c ^ 0x40);        /* Make it "^X"                 */
        col++;                  /* Count this char              */
    } else
        mlout(c);               /* Otherwise, output raw char   */
    TTflush();                  /* Flush the output             */
    return (++col);             /* return the new column number */
}



/*++

Simply clears the message line for macro usage

++*/
int clear_msg(f, n)
int f, n;
{

    f = n = 0;                  /* splint */

    mlforce("");
    return (TRUE);
}



/*++

This function writes a string on the message line for macro usage

++*/
int write_msg(f, n)
int f, n;
{
    int status;
    char buf[NPAT];             /* buffer to recieve message into */

    f = n = 0;                  /* splint */

    if ((status = mlreply(TEXT67, buf, NPAT - 1)) != TRUE)      /*"Message to write: " */
        return (status);

    /* expand all '%' to "%%" so mlwrite won't expect arguments */
    make_literal(buf);

    mlforce(buf);
    return (TRUE);
}
